package ics.uci.edu.Calico.UI.components;

import ics.uci.edu.Calico.UI.CalicoCanvas;
import ics.uci.edu.Calico.models.ObjectHandlerModel;
import ics.uci.edu.Calico.models.actions.ChangedCell;
import ics.uci.edu.Calico.models.actions.GridView;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.geom.Area;
import java.awt.geom.Point2D;
import java.awt.geom.RoundRectangle2D;
import java.awt.image.BufferedImage;
import java.io.Serializable;
import java.util.ArrayList;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

import edu.umd.cs.piccolo.nodes.PImage;
import edu.umd.cs.piccolo.util.PBounds;
import edu.umd.cs.piccolo.util.PPaintContext;
import edu.umd.cs.piccolox.nodes.PLine;
import edu.umd.cs.piccolo.event.PBasicInputEventHandler;
import edu.umd.cs.piccolo.event.PInputEvent;

/**
 * This class is a grid of VBGridCells, which in themselves have an image and an ObjectHandlerModel object.
 * There is a set number of cells that this can have, declared by MAX_ROWS and MAX_COLUMNS, but only a fraction
 * of these are intended to be visible initially. The visual matrix will resize itself if more columns are requested
 * but it will return bad values when the absolute walls are hit (getAdjacentCellPosition is where the resizing
 * of the visual grid happens).
 * 
 * Each grid cell has a position number (which can be obtained through get methods). The numbering of the positioning
 * is based off the theoretical max number of cells, so even if there is only a 4x4 matrix visible, the numbering
 * is based off the maximum grid size. This is done to preserve an individual cell's position number as the visual
 * grid increases.
 * 
 * @author nmangano
 *
 */
public class CalicoSharedGrid extends CalicoComponent {

	public static int NORTH = 0;
	public static int SOUTH = 1;
	public static int EAST = 2;
	public static int WEST = 3;
	
	public static final int MAX_ROWS = 64;
	public static final int MAX_COLUMNS = 64;
	
	//these four variables constitute the "Visual Grid". The visual grid has its upperleft corner specified by
	//row0,col0
	int row0, col0,
		numRows, numColumns;
	int currentCell;
	CalicoCanvas canvas;
	JFrame frame;
	
	/**
	 * @param row
	 * @param col
	 */
	public CalicoSharedGrid(int row, int col) {
		setGridSize(row, col);
		row0 = MAX_ROWS/2 - numRows/2;
		col0 = MAX_COLUMNS/2 - numColumns/2;
		currentCell = getStartingCell();
	
		this.addInputEventListener(new GridMouseListener());
		
	}

	public static int getStartingCell() {
		return (MAX_ROWS * MAX_COLUMNS)/2 - MAX_COLUMNS/2;
	}
	
	public void paint(PPaintContext context) {
		Graphics2D g2 = context.getGraphics();
		//g2.setPaint(Color.white);
		g2.setColor(new Color(230, 230, 230));
	
		Rectangle r = this.getBounds().getBounds();
		g2.fill(r);
	}
	
	public void clear() {
		removeAllChildren();
		readjustGrid();
	}
	
	/**
	 * Overrides the super class so the cells are repositioned and resized according to the window.
	 */
	public boolean setBounds(Rectangle rect) {	
		
		boolean ret = false;
		if (super.setBounds(rect))
			ret = true;
		readjustGrid();
		return ret;
	}
	
	/**
	 * Warning, this clears the matrix! This method creates the grid of lines on the matrix (which is a set
	 * of PLines).
	 * @param row
	 * @param col
	 */
	private void setGridSize(int row, int col) {
		removeAllChildren();
		
		numRows = row;
		numColumns = col;
		
		loadEmptyCells();
		//setNumRows(row);
		//setNumCols(col);
	}
	
	/**
	 * Called by setGridSize()
	 * @param r
	 */
	private void setNumRows(int r) {
		PLine line;
		//double cellHeight = new Double(getBounds().getHeight() / r).intValue();
		double cellHeight = getCellHeight();
		
		for (int i = 1; i < r; i++) {
			line = new PLine();
			line.addPoint(0, 0, cellHeight * i);
			line.addPoint(1, getBounds().getWidth(), cellHeight * i);
			line.setStroke(new BasicStroke(1.0f));
			addChild(line);
		}
	}
	
	/**
	 * Called by setGridSize()
	 * @param c
	 */
	private void setNumCols(int c) {
		PLine line;
		//double cellWidth = new Double(getBounds().getWidth() / c).intValue();
		double cellWidth = getCellWidth();
		
		for (int i = 1; i < c; i++) {
			line = new PLine();
			line.addPoint(0, cellWidth * i, 0);
			line.addPoint(1, cellWidth * i, getBounds().getHeight());
			line.setStroke(new BasicStroke(1.0f));
			addChild(line);
		}
	}
	
	/**
	 * This is the principle method to populate the matrix. If a cell already exists, it will be over-written.
	 * @param pos
	 * @param img
	 * @param m
	 */
	public void setCell(int pos, Image img, ObjectHandlerModel m) {		
		if (MAX_ROWS*MAX_COLUMNS < pos) {
			System.err.println("VBGrid:addImage() - pos of " + pos + " is bigger than the available cells " + MAX_ROWS*MAX_COLUMNS);
			return;
		}
		//remove the existing cell
		VBGridCell temp = getCellByPosition(pos);
		if (temp != null)
			removeChild(temp);
		
		VBGridCell image = new VBGridCell(pos);
		
		int height = new Double(getBounds().getHeight() / numRows).intValue();
		int width = new Double(getBounds().getWidth() / numColumns).intValue();
		if (img != null)
			image.setImage(CalicoCanvas.getProportionateScaledImage(img, width, height, Image.SCALE_SMOOTH));
		if (m != null)
			image.setModel(m);
		Rectangle bounds = getCellBounds(pos);
		
		//add it to the canvas
		addChild(image);
		image.setBounds(bounds);
		image.moveToBack();

	}
	
	/**
	 * This is a private function that readjusts the cell sizes and redraws the lines. The function is
	 * called when the size of the matrix is changed, or the window is resized.
	 *
	 */
	private void readjustGrid() {
		//grab all the existing cells
		ArrayList<VBGridCell> cells = new ArrayList<VBGridCell>();		
		for (int i = 0; i < getChildrenCount(); i++) {
			if (getChild(i) instanceof VBGridCell) {
				cells.add((VBGridCell)getChild(i));
				cells.get(i).setBounds(getCellBounds(cells.get(i).position));
			}
		}
		//reset all the lines on the grid (keep in mind this wipes everything!)
		setGridSize(numRows, numColumns);
		
		//put all the resized cells back on the layer (draw them last)
		for (VBGridCell c : cells) {
			this.setCell(c.position, c.getImage(), c.getModel());
		}
		
		repaint();		
	}
	
	/**
	 * Removes a cell based on its position
	 * @param pos
	 */
	public void removeImage(int pos) {
		if (numRows*numColumns < pos) {
			System.err.println("VBGrid:addImage() - pos of " + pos + " is bigger than the available cells " + numRows*numColumns);
			return;
		}
		
		//figure out where the point would be, and use the other algorithm
		Rectangle bounds = getCellBounds(pos);
		Point2D p = new Point2D.Double(bounds.getX()+1, bounds.getY()+1);
		removeImage(p);
	}
	
	
	/**
	 * Removes a cell based on its visual position.
	 * @param p
	 */
	public void removeImage(Point2D p) {
		for (int i = 0; i < getChildrenCount(); i++) {
			if (getChild(i) instanceof VBGridCell &&
					((VBGridCell)getChild(i)).getBounds().contains(p))
				removeChild(i);
			break;
		}
	}
	
	private void loadEmptyCells() {
		int pos;
		for (int x = 1, y = 1; getPositionByPoint(x,y) != -1; x += getCellWidth()) {
			for (y = 1; getPositionByPoint(x,y) != -1; y += getCellHeight()) {
				pos = getPositionByPoint(x,y);
				if (getCellByPosition(pos) == null) {
					//BufferedImage img = new BufferedImage(new Double(getBounds().width).intValue(), new Double(getBounds().height).intValue(), BufferedImage.TYPE_INT_RGB);
					//Graphics2D g = (Graphics2D)img.getGraphics();
					//g.setColor(Color.RED);
					//g.fillRect(0,0, new Double(getBounds().width).intValue(), new Double(getBounds().height).intValue());
					//VBGridMenu.saveImage(img, "EmptyCell.jpg");
					setCell(pos, null, null);
				}
			}
			y = 1;
		}
		
	}
	
	/**
	 * Please use getCellBounds instead.
	 * @param pos
	 * @return
	 */
	@Deprecated
	public Rectangle getPositionBounds(int pos) {
		return getCellBounds(pos);
	}
	
	/**
	 * Returns the bounds of a cell based on its position. If a cell is not in the visual grid, then
	 * null is returned.
	 * @param pos
	 * @return
	 */
	public Rectangle getCellBounds(int pos) {
		int x, y;
		
		Point2D coord = getCoordinate(pos);
		int row = new Double(coord.getX()).intValue();
		int col = new Double(coord.getY()).intValue();
		
		int cellHeight = getCellHeight();
		int cellWidth = getCellWidth();
		
		if (col < col0 || col > col0 + numColumns ||
			row < row0 || row > row0 + numRows )
			return null;
		
		x = cellWidth * (col - col0);
		y = cellHeight * (row - row0);
		
		return new Rectangle(x, y, cellWidth, cellHeight);
		
	}
	
	public int getPositionByPoint(int x, int y) {
		return getPositionByPoint(new Point2D.Double(x,y));
	}
	
	/**
	 * Returns a position based on its visual point.
	 * @param p
	 * @return
	 */
	public int getPositionByPoint(Point2D p) {
		
		int pos;
		
		for (int r = row0; r < row0 + numRows; r++) {
			for (int c = col0; c < col0 + numColumns; c++) {
				pos = (MAX_COLUMNS*(r)) + c;
				if (getCellBounds(pos).contains(p))
					return pos;
			}
		}
		
		return -1;
	}
	
	/**
	 * The principle method for getting cells. Returns a VBGridCell object based on its position.
	 * @param pos
	 * @return
	 */
	public VBGridCell getCellByPosition(int pos) {
		Rectangle bounds = getCellBounds(pos);
		if (bounds == null)
			return null;
		
		Point2D p = new Point2D.Double(bounds.getX() + 1, bounds.getY() + 1); 
		return getCellByPoint(p);
	}
	
	/**
	 *
	 * @return Returns a cell width that maintains a consistent ratio with the cell height.
	 */
	public int getCellWidth() {
		double scalingRatio = 1d / Math.max(numRows, numColumns);
		
		return new Double(getBounds().getWidth() * scalingRatio).intValue();
	}
	
	/**
	 * 
	 * @return Returns a cell height that maintains a consistent ratio with the cell width.
	 */
	public int getCellHeight() {
		double scalingRatio = 1d / Math.max(numRows, numColumns);
		
		return new Double(getBounds().getHeight() * scalingRatio).intValue();
	}
	
	/**
	 * Same as getCellByPosition, except it just returns the contained model.
	 * @param pos
	 * @return
	 */
	public ObjectHandlerModel getModelByPosition(int pos) {
		VBGridCell c = getCellByPosition(pos);
		if (c == null)
			return null;
		else
			return c.getModel();
	}
	
	/**
	 * 
	 * @param p
	 * @return The VBGrid cell based on the visual point.
	 */
	public VBGridCell getCellByPoint(Point2D p) {
		for (int i = 0; i < getChildrenCount(); i++) {
			if (getChild(i) instanceof VBGridCell &&
					((VBGridCell)getChild(i)).getBounds().contains(p))
				return (VBGridCell)getChild(i);
		}
		return null;
	}
	
	/**
	 * Returns the position number based on its position in the matrix.
	 * @param row
	 * @param col
	 * @return
	 */
	public int getPositionByRowCol(int row, int col) {
			return row*MAX_COLUMNS + col;
	}
	
	/**
	 * 
	 * @param pos The numerical position of the cell.
	 * @return Returns the row, column as a point (where X is row, and Y is the col).
	 */
	public Point2D getCoordinate(int pos) {
		
		int row = pos/MAX_COLUMNS;
		int col = pos - row*MAX_COLUMNS;
		
		return new Point2D.Double(row, col);
	}
	
	/**
	 * Returns the position of the adjacent cell in a specified direction. If a cell in the specific direction
	 * isn't visible, then the visible matrix is expanded. If we're at the very edge, then -1 is returned.
	 * 
	 * @param cell
	 * @param direction
	 * @return
	 */
	public int getAdjacentCellPosition(int cell, int direction) {
		//TODO: finish
		int row = cell/MAX_COLUMNS;
		int col = cell - row*MAX_COLUMNS;
		
		//figure out if it's out of bounds
		boolean outOfBounds = false;
		if (direction == NORTH) {
			if (row == row0)
				outOfBounds = true;
			row--;
		}
		else if (direction == SOUTH) {
			if (row == row0 + numRows - 1)
				outOfBounds = true;
			row++;
		}
		else if (direction == EAST) {
			if (col == col0 + numColumns - 1)
				outOfBounds = true;
			col++;
		}
		else if (direction == WEST) {
			if (col == col0)
				outOfBounds = true;
			col--;
		}
		//if it's out of bounds of the visual grid, readjust the visible grid.
		if (outOfBounds)
			if (direction == NORTH) {
				if (row0 == 0)
					return -1;
				else {
					//expand grid
					row0--;
					numRows++;
					readjustGrid();
				}
			}
			else if (direction == SOUTH) {
				if (row0 + numRows >= MAX_ROWS)
					return -1;
				else {
					//expand grid
					numRows++;
					readjustGrid();
				}
			}
			else if (direction == EAST) {
				if (col0 + numColumns >= MAX_COLUMNS)
					return -1;
				else {
					//expand grid
					numColumns++;
					readjustGrid();
				}
			}
			else if (direction == WEST) {
				if (col0 == 0)
					return -1;
				else {
					//expand grid
					col0--;
					numColumns++;
					readjustGrid();
				}
			}
			
		return getPositionByRowCol(row, col);
	}
	
	/**
	 * Copies a cell in the direction specified. If a cell exists in that direction, then that cell is moved one
	 * over in the same direction (this happens recursively).
	 * @param cell
	 * @param direction
	 * @return The position of the cell copied.
	 */
	public int copyCell(int cell, int direction) {
		VBGridCell origCell = getCellByPosition(cell);
		int dest = getAdjacentCellPosition(cell, direction);
		if (getCellByPosition(dest) != null &&
				!getCellByPosition(dest).isEmpty())
			copyCell(dest, direction);
		setCell(dest, origCell.getImage(), new ObjectHandlerModel(origCell.model));
		
		return dest;
	}
	
	private class GridMouseListener extends PBasicInputEventHandler {
		
		private static final int NO_ACTION = -1;
		private static final int TRANSLATE = 1;
		
		VBGridCell floater;
		PInputEvent mDown;
		
		int action = -1;
		
		public void mousePressed(PInputEvent e) {
			mDown = e;
			if (e.isRightMouseButton() && 
					getCellByPoint(e.getCanvasPosition()) != null) {
				action = TRANSLATE;
				floater = getCellByPoint(e.getCanvasPosition());
				floater.moveToFront();
			}
			
		}
		
		public void mouseReleased(PInputEvent e) {
			if (action == TRANSLATE) {
				double midX = floater.getFullBoundsReference().getX() + floater.getFullBoundsReference().getWidth()/2;
				double midY = floater.getFullBoundsReference().getY() + floater.getFullBoundsReference().getHeight()/2;
				//double midX = floater.getX() + floater.getWidth()/2;
				//double midY = floater.getY() + floater.getHeight()/2;
				
				int pos = getPositionByPoint(new Point2D.Double(midX, midY));
				if (pos != -1) {
					setCell(floater.position, null, null); 
					setCell(pos, floater.getImage(), floater.getModel());
					
				}
				else {
					floater.setBounds(getCellBounds(floater.position).getBounds());
				}
				repaint();
			}
		}
		
		public void mouseDragged(PInputEvent e) {
			if (action == TRANSLATE) {
				double dX = e.getCanvasDelta().getWidth();
				double dY = e.getCanvasDelta().getHeight();
				floater.translate(dX, dY);
			}
		}

		public void mouseClicked(PInputEvent e) {
			VBGridCell cell = getCellByPoint(e.getCanvasPosition());
			int position = getPositionByPoint(e.getCanvasPosition());
			if (cell != null)
				loadDrawingCanvas(position, cell.getModel());
			else if (position > -1) {
				ObjectHandlerModel emptyModel = new ObjectHandlerModel();
				loadDrawingCanvas(position, emptyModel);
			}		
		}
	}
	
	/**
	 * The principle unit within the matrix.
	 * @author nmangano
	 *
	 */
	public class VBGridCell extends PImage implements Serializable {
		
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		ObjectHandlerModel model;
		public int position;
		private Area border, shadow;
		
		public VBGridCell(int pos) {
			position = pos;
		}
		
		public boolean isEmpty() {
			return (model == null);
		}
		
		public void setModel(ObjectHandlerModel m) {
			model = m;
		}
		
		public ObjectHandlerModel getModel() {
			if (model == null)
				return new ObjectHandlerModel();
			else
				return model;
		}
		
		public void setImage(Image img) {
			Rectangle rect = new Rectangle(0, 0, img.getWidth(null), img.getHeight(null));
			RoundRectangle2D rRect = new RoundRectangle2D.Double(2, 2, img.getWidth(null) - 4, img.getHeight(null) - 4, 8, 8);
			Area border = new Area(rect);
			border.subtract(new Area(rRect));
			
			BufferedImage bimg = new BufferedImage(img.getWidth(null), img.getHeight(null), BufferedImage.TYPE_INT_ARGB);
			Graphics2D g = (Graphics2D)bimg.createGraphics();
			g.setComposite(AlphaComposite.Src);
			g.drawImage(img, 0, 0, null);
			g.setColor(Color.green);
			g.fill(border);
			g.setColor(new Color(100,100,100));
			g.draw(rRect);
			g.setColor(Color.black);
			g.drawRoundRect(new Double(getBounds().x).intValue() + 2, new Double(getBounds().y).intValue() + 2, 
					new Double(getBounds().width).intValue() - 4, new Double(getBounds().height).intValue() - 4, 10, 10);
			g.dispose();
			for(int i = 0; i < bimg.getHeight(); i++) {
				for(int j = 0; j < bimg.getWidth(); j++) {
					if(bimg.getRGB(j, i) == Color.green.getRGB()) {  
						bimg.setRGB(j, i, 0x8F1C1C);  
					}
				}
			}
			
			
			super.setImage(bimg);
			
		}
		
		public boolean setBounds(Rectangle r) {
			super.setBounds(r);
			Rectangle rect = new Rectangle(r.x, r.y, r.width, r.height);
			RoundRectangle2D rRect = new RoundRectangle2D.Double(r.x + 2, r.y + 2, r.width - 4, r.height - 4, 10, 10);
			border = new Area(rect);
			border.subtract(new Area(rRect));
			
			shadow = new Area(rRect);
			RoundRectangle2D rRect2 = new RoundRectangle2D.Double(r.x + 1, r.y + 1, r.width - 4, r.height - 4, 10, 10);
			shadow.subtract(new Area(rRect2));
			
			return true;
		}
		
		public void paint(PPaintContext c) {
			Image image = getImage();
			
			Graphics2D g2 = c.getGraphics();
			g2.setStroke(new BasicStroke(1.0f));
			g2.setColor(Color.black);
			//g2.drawRoundRect(new Double(getBounds().x).intValue() + 3, new Double(getBounds().y).intValue() + 3, 
			//		new Double(getBounds().width).intValue() - 4, new Double(getBounds().height).intValue() - 4, 10, 10);
			if (getImage() != null) {
				double iw = image.getWidth(null);
				double ih = image.getHeight(null);
				PBounds b = getBoundsReference();

				if (b.x != 0 || b.y != 0 || b.width != iw || b.height != ih) {
					g2.translate(b.x, b.y);
					g2.scale(b.width / iw, b.height / ih);
					g2.drawImage(image, 0, 0, null);
					g2.scale(iw / b.width, ih / b.height);
					g2.translate(-b.x, -b.y);
				} 
			}
			else {
				g2.setColor(Color.white);
				g2.fillRoundRect(new Double(getBoundsReference().x).intValue() + 2, new Double(getBoundsReference().y).intValue() + 2, 
						new Double(getBoundsReference().width).intValue() - 4, new Double(getBoundsReference().height).intValue() - 4,
						10, 10);
				g2.setColor(new Color(100,100,100));
				
				//g2.drawRoundRect(new Double(getBounds().x).intValue() + 2, new Double(getBounds().y).intValue() + 2, 
				//	new Double(getBounds().width).intValue() - 4, new Double(getBounds().height).intValue() - 4, 10, 10);
			}
				
			
			//g2.fill(border);
			
			//g2.draw(shadow);
			
		}
		
		
	}
	
	public void loadGrid() {
		canvas.fireListeners(new GridView(currentCell));
		canvas.getCamera().addChild(this);
		saveModelToGrid();
		//frame.setJMenuBar(new CalicoGridMenu(this));
		frame.pack();
		canvas.repaint();
	}
	
	public void loadDrawingCanvas(int pos, ObjectHandlerModel model) {
		canvas.getCamera().removeChild(this);
		canvas.loadModel(model);
		canvas.fireListeners(new ChangedCell(pos, currentCell));
		currentCell = pos;
		frame.setJMenuBar(null);
		frame.pack();
		canvas.repaint();
	}
	
	public void shiftCanvas(int direction, int action) {
		saveModelToGrid();
		int dest = -1;
		
		if (action == GridNavThumb.MOVE) {
			dest = getAdjacentCellPosition(currentCell, direction);
			ObjectHandlerModel model = getModelByPosition(dest);
			if (model == null)
				canvas.loadModel(new ObjectHandlerModel());
			else
				canvas.loadModel(model);
		}
		else if (action == GridNavThumb.COPY) {
			
			dest = copyCell(currentCell, direction);
			
			//we want to disassociate this model from the saved version
			canvas.loadModel(canvas.getModel());
			JOptionPane.showMessageDialog(null, "Your work as successful copied.");
		}
		canvas.fireListeners(new ChangedCell(dest, currentCell));
		currentCell = dest;
	}

	public void setCurrentGridPoint(int p) {
		currentCell = p;
	}
	
	public void saveModelToGrid() {
		setCell(currentCell, canvas.getImage(canvas.getBackground()), canvas.getModel());
	}
	
	public void enableMenu(JFrame f) {
		frame = f;
	}

	@Override
	public void registerCanvas(CalicoCanvas c) {
		canvas = c;
		//We only want this present situationally
		canvas.getCamera().removeChild(this);
		//canvas.addComponent(new GridNavPointer));
		canvas.addComponent(new GridNavThumb(GridNavThumb.NORTH));
		canvas.addComponent(new GridNavThumb(GridNavThumb.SOUTH));
		canvas.addComponent(new GridNavThumb(GridNavThumb.EAST));
		canvas.addComponent(new GridNavThumb(GridNavThumb.WEST));
	}

	@Override
	public boolean updateBounds(int width, int height) {		
		if (setBounds(new Rectangle(0, 0, width, height)))
			return true;
		return false;
	}
	
	
	
	
}